import tensorflow as tf
import keras

@keras.saving.register_keras_serializable()
class BoundaryAwareDiceLoss(tf.keras.losses.Loss):
    """Boundary aware Dice Loss 
    
    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    def __init__(self, alpha=1, beta=1, gamma=1, epsilon=1e-5, **kwargs):
        super(BoundaryAwareDiceLoss, self).__init__(**kwargs)
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.epsilon = epsilon

    def call(self, y_true, y_pred):
        intersection = tf.reduce_sum(y_true * y_pred)
        union = tf.reduce_sum(y_true) + tf.reduce_sum(y_pred)
        dice_loss = 1 - (2 * intersection + self.beta) / (union + self.alpha + self.gamma + self.epsilon)
        return dice_loss

# Example usage
# loss_function = BoundaryAwareDiceLoss(alpha=1, beta=1, gamma=1, epsilon=1e-5)

# Assuming y_true and y_pred are your ground truth and predicted masks, respectively
# loss_value = loss_function(y_true, y_pred)
# loss_function = BoundaryAwareDiceLoss(alpha=1, beta=1, gamma=1, epsilon=1e-5)

def custom_BAD_loss(y_true, y_pred):
    BAD_loss = BoundaryAwareDiceLoss(alpha=1, beta=1, gamma=1, epsilon=1e-5)
    return 0.01*tf.keras.metrics.binary_crossentropy(y_true, y_pred) + 0.99*BAD_loss(y_true, y_pred)


# Train test split
import tensorflow as tf
# import os
# import numpy as np
# from tqdm import tqdm 
# from skimage.io import imread, imshow
#from skimage.transform import resize
from sklearn.model_selection import train_test_split
from keras.utils import to_categorical
# import matplotlib.pyplot as plt


def get_train_test_val_iterators(
    X, Y, 
    batch_size=20, 
    train_test_ratio=0.2, 
    train_val_ratio=0.1, 
    train_test_seed=42, 
    train_val_seed=42):
    """Prepare Train, Test and Validation data iterators

    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """

    X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=train_test_ratio, random_state=train_test_seed)
    print(X_train.shape==Y_train.shape, X_train.shape)
    del X, Y
    print(X_train.shape, Y_train.shape, X_train.dtype, Y_train.dtype, type(X_train), type(Y_train), X_test.shape, Y_test.shape, X_test.dtype, Y_test.dtype, type(X_test), type(Y_test))


    # Validation split
    # from sklearn.model_selection import train_test_split
    X_train, X_val, Y_train, Y_val = train_test_split(X_train, Y_train, test_size=train_val_ratio, random_state=train_val_seed) # 0.25 x 0.8 = 0.2
    print('Train ', X_train.shape==Y_train.shape, X_train.shape, X_train.dtype==Y_train.dtype, X_train.dtype)
    print('Val   ', X_val.shape==Y_val.shape, X_val.shape, X_val.dtype==Y_val.dtype, X_val.dtype)
    print('Test. ', X_test.shape==Y_test.shape, X_test.shape, X_test.dtype==Y_test.dtype, X_test.dtype)



    # Normalization (assumned inside model)
    datagen = tf.keras.preprocessing.image.ImageDataGenerator(...)
    datagen.fit(X_train)



    ## new categorical
    # Assuming Y_train is a 1D array containing class labels
    # Convert Y_train to categorical format
    Y_train_categorical = to_categorical(Y_train)
    del Y_train
    Y_val_categorical = to_categorical(Y_val)
    del Y_val
    Y_test_categorical = to_categorical(Y_test)
    del Y_test

    # batch_size = 20
    train_iterator = datagen.flow(X_train, Y_train_categorical, batch_size=batch_size)
    val_iterator = datagen.flow(X_val, Y_val_categorical, batch_size=batch_size)
    test_iterator = datagen.flow(X_test, Y_test_categorical, batch_size=batch_size)
    # del to_categorical

    return train_iterator, test_iterator, val_iterator

# train_iterator, test_iterator, val_iterator = get_train_test_val_iterators(X,Y)


import tensorflow as tf
import tensorflow.keras.backend as K

# # tf 2.15
# def dice_coef(y_true, y_pred, smooth=0):
#     y_true_f = K.flatten(y_true)
#     y_pred_f = K.flatten(y_pred)
#     intersection = K.sum(y_true_f * y_pred_f)
#     dice = (2. * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)
#     return dice

# tf 2.16
# import tensorflow.keras.backend as K
def dice_coef(y_true, y_pred, smooth=0):
    """Dice coefficient 
    
    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    # print(y_true.dtype, y_pred.dtype)
    y_true = tf.cast(y_true, tf.float32)
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    dice = (2. * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)
    return dice


import time
import datetime

# from datetime import timedelta
def elapsedtime(start_time, end_time):
    """Compute elapsed time

    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    # Calculate elapsed time
    elapsed_time_seconds = end_time - start_time

    # Convert seconds to a timedelta object
    delta = datetime.timedelta(seconds=elapsed_time_seconds)

    # Extract minutes, seconds, and milliseconds from the timedelta object
    minutes = delta.seconds // 60
    seconds = delta.seconds % 60
    milliseconds = delta.microseconds // 1000

    # Format the result
    formatted_time = "{:02}:{:02}:{:03}".format(minutes, seconds, milliseconds)
    print(f"Elapsed time: mm:ss:mss \t {formatted_time}")
    return formatted_time

## Somewhere above

# start_time=time.time()

## Time elapsed
# end_time = time.time()
# elapsedtime(start_time, end_time)



## time stamp
# import time
# import datetime
def timestamp(time):
    """Create a time stamp

    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    # Get the current timestamp using time.time()
    timestamp = time

    # Convert the timestamp to a datetime object
    dt_object = datetime.datetime.fromtimestamp(timestamp)

    # Print the datetime object
    print(dt_object)
    # formatted_string = dt_object.strftime('%Y%m%d%H%M%S')
    formatted_string = dt_object.strftime('%Y%m%d')
    return formatted_string

# timestamp(time.time())


""" MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import tensorflow as tf

# Unet configurations
uconfigs = {
    '12345':(2,4,8,16,32),
    '23456':(4,8,16,32,64),
    '34567':(8,16,32,64,128),
    '45678':(16,32,64,128,256),
    '56789':(32,64,128,256,512)
    }    

#%%    
def Unet2D(
    n_classes=2, 
    n_input_channels=1, 
    scale=1, 
    input_shape =(256, 256, 1),
    # dataset='IBSR', 
    # config=configs['12345'],
    config=uconfigs['45678'],
    loss = 'binary_crossentropy',
    optimizer = tf.keras.optimizers.Adam(learning_rate=0.001),
    residual=False,
    compile=True):
    """UNet2D (control model) 
    
    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    """

    a, b, c, d, e = config
    
    #inputs = tf.keras.layers.Input((IMG_HEIGHT, IMG_WIDTH, IMG_CHANNELS))
    # inputs = tf.keras.layers.Input((256, 256, 1))
    # s=inputs
    inputs = tf.keras.layers.Input(input_shape)
    # s = tf.keras.layers.Resizing(
    #                 height=256,
    #                 width=256,
    #                 interpolation='bilinear',
    #                 crop_to_aspect_ratio=False
    #             )(inputs)


    s = inputs
    s = tf.keras.layers.Lambda(lambda x: x / 255)(s)
    # s = tf.keras.layers.Lambda(lambda x: tf.where(
    #     tf.reduce_sum(x) !=0,
    #     (((x - tf.reduce_mean(x)) / tf.math.reduce_std(x)) - (tf.reduce_min((x - tf.reduce_mean(x)) / tf.math.reduce_std(x)))) 
    #    / ((tf.reduce_max((x - tf.reduce_mean(x)) / tf.math.reduce_std(x))) - (tf.reduce_min((x - tf.reduce_mean(x)) / tf.math.reduce_std(x)))), 
    #     x
    #     ))(s)
    #s = Normlayer(inputs_resiz)

    #s = tf.keras.layers.Lambda(lambda x: x / 255)(inputs)

    #a,b,c,d,e=2,4,8,16,32
    # a,b,c,d,e=4,8,16,32,64 ##it was here
    # a,b,c,d,e=8,16,32,64,128
    # a,b,c,d,e=16,32,64,128,256
    #Contraction path
    c1 = tf.keras.layers.Conv2D(a, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(s)
    c1 = tf.keras.layers.Dropout(0.1)(c1)
    c1 = tf.keras.layers.Conv2D(a, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c1)
    p1 = tf.keras.layers.MaxPooling2D((2, 2))(c1)

    c2 = tf.keras.layers.Conv2D(b, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(p1)
    c2 = tf.keras.layers.Dropout(0.1)(c2)
    c2 = tf.keras.layers.Conv2D(b, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c2)
    p2 = tf.keras.layers.MaxPooling2D((2, 2))(c2)
    
    c3 = tf.keras.layers.Conv2D(c, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(p2)
    c3 = tf.keras.layers.Dropout(0.2)(c3)
    c3 = tf.keras.layers.Conv2D(c, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c3)
    p3 = tf.keras.layers.MaxPooling2D((2, 2))(c3)
    
    c4 = tf.keras.layers.Conv2D(d, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(p3)
    c4 = tf.keras.layers.Dropout(0.2)(c4)
    c4 = tf.keras.layers.Conv2D(d, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c4)
    p4 = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(c4)
    
    #bottom
    c5 = tf.keras.layers.Conv2D(e, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(p4)
    c5 = tf.keras.layers.Dropout(0.3)(c5)
    c5 = tf.keras.layers.Conv2D(e, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c5)

    #Expansive path 
    u6 = tf.keras.layers.Conv2DTranspose(d, (2, 2), strides=(2, 2), padding='same')(c5)
    u6 = tf.keras.layers.concatenate([u6, c4])
    c6 = tf.keras.layers.Conv2D(d, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(u6)
    c6 = tf.keras.layers.Dropout(0.2)(c6)
    c6 = tf.keras.layers.Conv2D(d, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c6)
    
    u7 = tf.keras.layers.Conv2DTranspose(c, (2, 2), strides=(2, 2), padding='same')(c6)
    u7 = tf.keras.layers.concatenate([u7, c3])
    c7 = tf.keras.layers.Conv2D(c, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(u7)
    c7 = tf.keras.layers.Dropout(0.2)(c7)
    c7 = tf.keras.layers.Conv2D(c, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c7)
    
    u8 = tf.keras.layers.Conv2DTranspose(b, (2, 2), strides=(2, 2), padding='same')(c7)
    u8 = tf.keras.layers.concatenate([u8, c2])
    c8 = tf.keras.layers.Conv2D(b, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(u8)
    c8 = tf.keras.layers.Dropout(0.1)(c8)
    c8 = tf.keras.layers.Conv2D(b, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c8)
    
    u9 = tf.keras.layers.Conv2DTranspose(a, (2, 2), strides=(2, 2), padding='same')(c8)
    u9 = tf.keras.layers.concatenate([u9, c1], axis=3)
    c9 = tf.keras.layers.Conv2D(a, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(u9)
    c9 = tf.keras.layers.Dropout(0.1)(c9)
    c9 = tf.keras.layers.Conv2D(a, (3, 3), activation='relu', kernel_initializer='he_normal', padding='same')(c9)

    # out_resiz  = tf.keras.layers.Resizing(
    #                 height=256,
    #                 width=192,
    #                 interpolation='bilinear',
    #                 crop_to_aspect_ratio=False
    #             )(c9)

    # out_resiz=c9
    # outputs = tf.keras.layers.Conv2D(1, (1, 1), activation='sigmoid')(out_resiz)

    R = c9
    if residual==True:
        R = Concatenate()([s, R])
    R = tf.keras.layers.Conv2D(n_classes, (1, 1), activation="softmax")(R)
    outputs = R

     
    model = tf.keras.Model(inputs=[inputs], outputs=[outputs])


    if compile==True:
        # model = tf.keras.Model(inputs=[inputs], outputs=[x])
        # generator_optimizer = tf.keras.optimizers.Adam(learning_rate=2e-4)
        # model.compile(optimizer=generator_optimizer, loss=generator_loss)
       
        model.compile(optimizer=optimizer, loss=loss)
        return model
        #
    else: 
        return model


if __name__=='__main__':
    # Loss = BoundaryAwareDiceLoss(alpha=1, beta=1, gamma=1, epsilon=1e-5) ## BAD loss
    # generator_loss = 'binary_crossentropy'
    # generator_optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
    Unet2D(compile=False).summary()
    Unet2D(input_shape=(128,128,1), config=uconfigs['45678'], residual=True, compile=False).summary()
    # Gφψ(compile=False).summary()


"""ॐ
    Version 0.0.1: 2D MEDCNN without attentions

    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """



# import tensorflow as tf
# import keras
# import numpy as np

# from TFDWT.dbFBimpulseResponse import FBimpulseResponses
# from TFDWT.GETDWTFiltersOrtho import GETDWTFiltersOrtho
# from TFDWT.get_A_matrix_dwt_analysisFB_unit import get_A_matrix_dwt_analysisFB_unit

# # from TFDWT3D.DWTIDWT1Dv1 import DWT1D, IDWT1D
# # from TFDWT3D.DWTIDWT2Dv1 import DWT2D, IDWT2D





__version__="0.0.1"


# # # include ../dirx 
# mylibpath = [
#       '/home/kishoretarafdar/src/MEDCNN_copy'
#     ]
# import sys
# [sys.path.insert(0,_) for _ in mylibpath]
# del mylibpath

import tensorflow as tf
# from BoundaryAwareDiceLoss import BoundaryAwareDiceLoss
# from dice import dice_coef
from keras.utils import plot_model 

def compile_model(model, dataset, dice_coef):
    """Compile model 
    
    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    if dataset == 'IBSR':
        # Loss, lossname = BoundaryAwareDiceLoss(alpha=1, beta=1, gamma=1, epsilon=1e-5), 'bad' ## BAD loss
        Loss, lossname = 'binary_crossentropy', 'bce'
    elif dataset == 'ATLAS':
        Loss, lossname = 'binary_crossentropy', 'bce'

    optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
    iou_metric = tf.keras.metrics.BinaryIoU(target_class_ids=(0,1), name='IoU')

    model.compile(optimizer=optimizer, loss=Loss, metrics=['accuracy', dice_coef, iou_metric], jit_compile=False)
    model.summary()
    plot_model(model, 
        show_shapes=True, 
        expand_nested=False, 
        show_layer_activations=True)
        # #   to_file=f'{wave}_Unet_{round(time.time(),4)}.png')
    return model, lossname


""" Multiresolution Encoder-Decoder Convolutional Neural Network (MEDCNN) without attentions
    Copyright (C) 2025 Kishore Kumar Tarafdar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""


# # include ../dirx 
# mylibpath = [
#     # '/home/kishor/src/FastDWTConvLayers',
#     '/home/kishor/src/MRSegmentation/Attentions19102023'
#     #'/home/k/PLAYGROUND10GB/SKULSTRIPpaper__'
#     ]
# import sys
# [sys.path.insert(0,_) for _ in mylibpath]
# del mylibpath

from TFDWT.DWTIDWT2Dv1 import DWT2D, IDWT2D
from tensorflow.keras.layers import Concatenate

import tensorflow as tf


#%% Pooling layers
# @keras.saving.register_keras_serializable()
class Pooling(tf.keras.layers.Layer):
    """DWT Pooling Layer: keep Low freq band only
   
    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    def __init__(self, Ψ='haar', **kwargs):
        super(Pooling, self).__init__(**kwargs)
        self.supports_masking = True
        self.Ψ = Ψ 

    #  def build(self, input_shape):
    #     self.num_channels = input_shape[-1]
    #     super(Pooling, self).build(input_shape) 

    def call(self, inputs):
        """inputs -> wave0 -> wave1 -> wave2 -> wave0_cap(inverse)"""
        chans = inputs.shape[3]
        wave0 = inputs #L0
        # wave1 = WaveTFFactory.build(self.Ψ)(wave0) #L1
        wave1 = DWT2D(self.Ψ)(wave0)
        return wave1[:,:,:,:chans]
        
#%%
configs = {
    # 'minimal': (2, 2, 2, 2, 2),
    'minimal2': (2, 4, 8, 16, 32),

    # 'config242': (2, 2, 2, 2, 4),
    # 'config282': (2, 2, 2, 4, 4),
    # 'config242': (2, 2, 4, 4, 4),
    # 'config282': (2, 4, 4, 4, 4),
    # 'config282': (2, 4, 4, 4, 8),
    # 'config282': (2, 4, 4, 8, 8),
    # 'config2F2': (2, 4, 8, 8, 8),
    # 'config282': (2, 4, 8, 16, 16),



    # 'config242': (4, 4, 4, 4, 4),
    # 'config282': (4, 4, 4, 4, 8),
    # 'config282': (4, 4, 4, 8, 8),
    # 'config282': (4, 4, 8, 8, 8),
    # 'config242': (4, 8, 8, 8, 8),
    # 'config282': (8, 8, 8, 8, 8),
    # 'config282': (8, 8, 8, 8, 16),
    # 'config2F2': (8, 8, 8, 16, 16),
    # 'config2F2': (8, 8, 16, 16, 16),
    # 'config2F2': (8, 8, 16, 16, 16),
    # 'config2F2': (8, 16, 16, 16, 16),
    # 'config2F2': (16, 16, 16, 16, 16),
    # 'config2F2': (2, 4, 8, 16, 32),
    # 'config2F2': (2, 4, 8, 16, 16),
    # 'config2F2': (2, 4, 8, 16, 16),
    
    
    # 'config4': (4, 4, 4, 4, 4),
    # 'config484': (4, 8, 8, 8, 4),
    # 'config4F4': (4, 16, 16, 16, 4),
    # 'config8': (8, 8, 8, 8, 8),
    # 'default': (8, 16, 16, 16, 8),
    # 'default2': (8, 16, 32, 16, 8)
}


#%%
def conv2D(f, x, activation='relu'):
    """Double conv layer"""
    x = tf.keras.layers.Conv2D(f, (3, 3), kernel_initializer='he_normal', padding='same')(x)
    # x = tf.keras.layers.BatchNormalization()(x)
    # x = tf.keras.layers.Activation(activation)(x)
    x = tf.keras.layers.Dropout(0.1)(x)
    x = tf.keras.layers.Conv2D(f, (3, 3), kernel_initializer='he_normal', padding='same')(x)
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.layers.Activation(activation)(x)
    # x = CBAM(reduction_ratio=2)(x)
    return x


N_CLASSES = 2 # Mutliclass
N_INPUT_CHANNELS = 1 # tens
# wave = 'haar'

#from keras.engine.training import h5py
# def dwtunet(dataset='IBSR', config=configs['config82']):
def Gφψ(
    Ψ='haar', 
    n_classes=N_CLASSES, 
    n_input_channels=N_INPUT_CHANNELS, 
    scale=1, 
    input_shape =(256, 256, 1),
    # dataset='IBSR', 
    config=configs['minimal2'],
    loss = 'binary_crossentropy',
    optimizer = tf.keras.optimizers.Adam(learning_rate=0.001),
    residual=False,
    compile=True):
    """ Residual MEDCNN without attention for binary semantic segmentation
   
    Multiresolution features:
        1. DWT level 4 decompostion of input matrix
        2. Wavelet pooling

    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """

    activation, activation2, activation3 = 'relu', 'relu', tf.keras.activations.mish
    # tfa.activations.mish


    # input_shape =(256, 256, 1)
    inputs = tf.keras.layers.Input(input_shape)
    # s  = tf.keras.layers.Resizing(
    #                 height=256,
    #                 width=256,
    #                 interpolation='bilinear',
    #                 crop_to_aspect_ratio=False
    #             )(inputs)    
    
    s = inputs
    
    # s = tf.keras.layers.Lambda(lambda x: x / 255)(s)
    # normalize
    s = tf.keras.layers.Lambda(lambda x: tf.where(
        tf.reduce_sum(x) !=0,
        (((x - tf.reduce_mean(x)) / tf.math.reduce_std(x)) - (tf.reduce_min((x - tf.reduce_mean(x)) / tf.math.reduce_std(x)))) 
       / ((tf.reduce_max((x - tf.reduce_mean(x)) / tf.math.reduce_std(x))) - (tf.reduce_min((x - tf.reduce_mean(x)) / tf.math.reduce_std(x)))), 
        x
        ))(s)
 
    
    l = DWT2D(wave=Ψ)(s)
    l1 = l[:,:,:,:1]
    h1 = l[:,:,:,1:]
    # print(l1.shape)
    #
    l = DWT2D(wave=Ψ)(l1)
    l2 = l[:,:,:,:1]
    h2 = l[:,:,:,1:]
    #
    l = DWT2D(wave=Ψ)(l2)
    l3 = l[:,:,:,:1]
    h3 = l[:,:,:,1:]
    #
    l = DWT2D(wave=Ψ)(l3)
    l4 = l[:,:,:,:1]
    h4 = l[:,:,:,1:]
    #
    
    # f, a, b, c, d = 8, 16, 16, 16, 8
    # f, a, b, c, d = config
    a, b, c, d , e = config


    #LP process
    # xl1_ = conv2D(f, l1)
    xl1_ = conv2D(a, l1)
    # xl1 = tf.keras.layers.MaxPooling2D((2, 2))(xl1_)
    xl1 = Pooling()(xl1_)    

    # xl2 = conv2D(f, l2)
    # xl2 = l2
    xl2 = Concatenate()([xl1, l2])
    # xl2 = tf.keras.layers.Conv2D(a, (3, 3), activation=activation, kernel_initializer='he_normal', padding='same')(xl2)
    # xl2_ = conv2D(a, xl2)
    xl2_ = conv2D(b, xl2)
    xl2 = Pooling()(xl2_)   

    # xl3 = conv2D(f, l3)
    # xl3 = l3
    xl3 = Concatenate()([xl2, l3])
    # xl3 = tf.keras.layers.Conv2D(b, (3, 3), activation=activation, kernel_initializer='he_normal', padding='same')(xl3)
    # xl3_ = conv2D(b, xl3)
    xl3_ = conv2D(c, xl3)
    xl3 = Pooling()(xl3_)
    
    # xl4 = conv2D(f, l4)
    # xl4 = l4
    xl4 = Concatenate()([xl3, l4])
    # xl4 = tf.keras.layers.Conv2D(c, (3, 3), activation=activation, kernel_initializer='he_normal', padding='same')(xl4)
    # xl4_ = conv2D(c, xl4)
    xl4_ = conv2D(d, xl4)
    xl4 = Pooling()(xl4_)


    xlat = xl4
    # e = f
    # floor------------
    xlat = conv2D(e, xlat)
    #-----------------    


    ## Decoder /Synthesis -----------------

    xl4u_ = tf.keras.layers.Conv2DTranspose(d, (2, 2), strides=(2, 2), padding='same')(xlat)
    xl4u_ = Concatenate()([xl4_, xl4u_])
    # xl4u_ = conv2D(c, xl4u_)
    xl4u_ = conv2D(d, xl4u_)
    xl4u = xl4u_    


    #xl4u_ #up
    #xl4u_ pass
    
    xl3u_ = tf.keras.layers.Conv2DTranspose(c, (2, 2), strides=(2, 2), padding='same')(xl4u_)
    xl3u_ = Concatenate()([xl3_, xl3u_])
    # xl3u_ = conv2D(b, xl3u_)
    xl3u_ = conv2D(c, xl3u_)
    xl3u = xl3u_    


    #xl3u_ #up
    #xl3u_ pass
  


    xl2u_ = tf.keras.layers.Conv2DTranspose(b, (2, 2), strides=(2, 2), padding='same')(xl3u_)
    xl2u_ = Concatenate()([xl2_, xl2u_])
    # xl2u_ = conv2D(a, xl2u_)
    xl2u_ = conv2D(b, xl2u_)
    xl2u = xl2u_    


    #xl2u_ #up
    #xl2u_ pass
  

    xl1u_ = tf.keras.layers.Conv2DTranspose(a, (2, 2), strides=(2, 2), padding='same')(xl2u_)
    xl1u_ = Concatenate()([xl1_, xl1u_])
    # xl1u_ = conv2D(f, xl1u_)
    xl1u_ = conv2D(a, xl1u_)
    xl1u = xl1u_   


    k = 3  
   
    iw4 = IDWT2D(wave=Ψ)(Concatenate()([xl4u,h4]))
    #
    iw3 = tf.keras.layers.Conv2D(1, (k, k), activation=activation3, padding='same')(Concatenate()([xl3u,iw4]))
    # iw3 = SpatialAttention()(Concatenate()([xl3u, l3, iw4]))
    iw3 = IDWT2D(wave=Ψ)(Concatenate()([iw3,h3]))
    #
    iw2 = tf.keras.layers.Conv2D(1, (k, k), activation=activation3, padding='same')(Concatenate()([xl2u,iw3]))
    # iw2 = SpatialAttention()(Concatenate()([xl2u, l2, iw3]))
    iw2 = IDWT2D(wave=Ψ)(Concatenate()([iw2,h2]))
    #
    iw1 = tf.keras.layers.Conv2D(1, (k, k), activation=activation3, padding='same')(Concatenate()([xl1u,iw2]))
    # iw1 = SpatialAttention()(Concatenate()([xl1u, l1, iw2]))
    R = IDWT2D(wave=Ψ)(Concatenate()([iw1,h1]))    
   
    
    
    ## OLD
    # R = tf.keras.layers.Conv2D(1, (1, 1), activation='sigmoid')(R)  
    
    ## NEW
    # s = tf.keras.layers.Conv2D(1, (1, 1), activation='sigmoid')(s)
    if residual==True:
        R = Concatenate()([s, R])
    
    R = tf.keras.layers.Conv2D(n_classes, (1, 1), activation="softmax")(R)

    outputs = R

     
    model = tf.keras.Model(inputs=[inputs], outputs=[outputs])

    if compile==True:
        # model = tf.keras.Model(inputs=[inputs], outputs=[x])
        # generator_optimizer = tf.keras.optimizers.Adam(learning_rate=2e-4)
        # model.compile(optimizer=generator_optimizer, loss=generator_loss)
       
        model.compile(optimizer=optimizer, loss=loss)
        return model
        #
    else: 
        return model


if __name__=='__main__':
    # Loss = BoundaryAwareDiceLoss(alpha=1, beta=1, gamma=1, epsilon=1e-5) ## BAD loss
    # generator_loss = 'binary_crossentropy'
    # generator_optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
    Gφψ(compile=False).summary()
    Gφψ(input_shape=(128,128,1), residual=True, compile=False).summary()
    # Gφψ(compile=False).summary()


import tensorflow as tf

def select_a_gpu(gpus:list, gpu_id:int, memory_limit=47):
    """Hard limit: 47 GB max gpu memory

    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    _select_gpu = gpus[gpu_id]
    if gpus:
        try:
            """limit memory to the seleced GPU"""
            # _limitGB = 32
            memory_limit = int(memory_limit*1024)
            tf.config.set_logical_device_configuration(
                _select_gpu, 
                [tf.config.LogicalDeviceConfiguration(memory_limit=memory_limit)])

            """Restrict tensorflow to use 1 GPU"""
            tf.config.set_visible_devices(_select_gpu, 'GPU')
            one_logical_gpu = tf.config.list_logical_devices('GPU')


            print(f"{len(gpus)} Physical GPUs available \nSelected {len(one_logical_gpu)} Logical GPU with {int(memory_limit/1024)} GB memory limit")
        except RuntimeError as e:
            # Visible devices must be set before GPUs have been initialized
            print(e)



if __name__=='__main__':
    gpus = tf.config.list_physical_devices('GPU')
    print("Num GPUs Available: ", len(gpus))  
    select_a_gpu(gpus, gpu_id=1, memory_limit=1)
    del select_a_gpu, gpus



"""Other examples: Cluster of virtual gpus in a GPU chip

    gpus = tf.config.list_physical_devices('GPU')
    _select_gpu = gpus[gpu_id]
    tf.config.set_logical_device_configuration(
        _select_gpu,
        [tf.config.LogicalDeviceConfiguration(memory_limit=1024),
            tf.config.LogicalDeviceConfiguration(memory_limit=1024)])
"""


# IBSR_allplanes_shuffled_resized_INPUTS_SS.npy 
# IBSR_allplanes_shuffled_resized_TARGETS_SS.npy
import os
import numpy as np

def load_ibsr_XY(datapath):
    """Load IBSRv2 data
    
    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    dataset = 'IBSR'
    # datapath = '/home/k/LOTUS/DATAMRI/IBSR_cooked_XY_SkullStripping'
    # datapath = '/home/kishor/FILESHARE/IBSR_cookedXY_allplanes/IBSR_XY_skullstripping/'
    # datapath = '/data1/kishoretarafdar/FILESHARE/IBSR_cookedXY_allplanes/IBSR_XY_skullstripping'
    # with open('IBSR_allplanes_shuffled_resized_INPUTS_SS.npy', 'rb') as f:
    with open(os.path.join(datapath, 'IBSR_X.npy'), 'rb') as f:
        X = np.load(f)
    del f
    # with open('IBSR_allplanes_shuffled_resized_TARGETS_SS.npy', 'rb') as f:
    with open(os.path.join(datapath, 'IBSR_Y.npy'), 'rb') as f:
        Y = np.load(f)
    del f#, datapath
    # X.shape, Y.shape
    return X, Y, dataset


def load_atlas_XY(datapath):
    """Load ATLAS2 data 
    
    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    dataset = 'ATLAS'
    with open(os.path.join(datapath, 'ATLAS_X_minnorm40.npy'), 'rb') as f:
        X = np.load(f)
    del f 
    with open(os.path.join(datapath, 'ATLAS_Y_minnorm40_int8.npy'), 'rb') as f:
        Y = np.load(f)
    del f
    # X.shape, Y.shape 
    return X, Y, dataset


# datasetname = 'IBSR'
# datasetname = 'ATALS'

# # include ../dirx 
mylibpath = [
    # '/home/kishor/src/FastDWTConvLayers',
    # '/home/kishor/src/MRSegmentation/Attentions19102023'
      # '/home/k/src/_MEDCNNsrc_part/MEDCNN_copy'
      '/home/kishoretarafdar/src/MEDCNN_copy'
      
    #'/home/k/PLAYGROUND10GB/SKULSTRIPpaper__'
    ]
import sys
[sys.path.insert(0,_) for _ in mylibpath]
del mylibpath

from utils import elapsedtime, timestamp
import shelve
import pickle
import matplotlib.pyplot as plt
import tensorflow as tf
import time

def train(model, train_iterator, test_iterator, val_iterator, dataset='IBSR', segconfig='nonResidual' , lossname='bce', CONFIGKEY='4567', epochs=40):
    """Train model

    MEDCNN: Multiresolution Encoder-Decoder Convolutional Neural Network
    Copyright (C) 2025 Kishore Kumar Tarafdar
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    """
    start_time = time.time()
    # MODEL_FNAME_PATTERN = f'{timestamp(start_time)}Ghaar_{datasetname}_{segconfig}_{lossname}_config{configkey}.keras'
    MODEL_FNAME_PATTERN = f'{timestamp(start_time)}G_{dataset}_{segconfig}_{lossname}_config{CONFIGKEY}.keras'
    # MODEL_FNAME_PATTERN = 'abc.keras'
    # MODEL_FNAME_PATTERN = f'{timestamp(start_time)}Grbio13_{datasetname}_{segconfig}_{lossname}_config{configkey}.keras'
    # model = dwtunet()
    # epochs =  40 # tot  = 120
    patience = epochs//2
    ###############################
    #Modelcheckpoint
    # checkpointer = tf.keras.callbacks.ModelCheckpoint('model_for_sagittal.h5', verbose=1, save_best_only=True)


    # patience = 10
    # patience = epochs #!!override early stopping
    callbacks = [
            tf.keras.callbacks.EarlyStopping(patience=patience, monitor='val_loss'),
            tf.keras.callbacks.ModelCheckpoint(filepath=MODEL_FNAME_PATTERN, verbose=1, save_best_only=True),
    ]
            # tf.keras.callbacks.TensorBoard(log_dir='dwtlogs')]

    #results = model.fit(X_train_resized, Y_train_resized, validation_split=0.1, batch_size=100, epochs=25, callbacks=callbacks)
    #results = model.fit(X_train, Y_train, validation_split=0.1, batch_size=20, epochs=1, callbacks=callbacks)
    results = model.fit(train_iterator, 
                        validation_data = val_iterator,
                        epochs=epochs, 
                        callbacks=callbacks)


    ####################################
    #TEST
    test_results = model.evaluate(test_iterator)#, batch_size=20)
    print("\n TEST:\ntest loss, test acc:", test_results)

    ## Show model weights distribution
    allp = [tf.reshape(_, [-1]).numpy() for _ in model.weights]
    len(allp)
    plt.figure(figsize=(60,8))
    plt.boxplot(allp)
    plt.show()
    del allp


    ## Time elapsed
    # end_time = time.time()
    elapsedtime(start_time, end_time=time.time())


    with shelve.open(f'G{CONFIGKEY}') as f:
        # Access variables stored in the shelve file
        f['train_result'] = results.history
        f['test_result'] = test_results

    
    """Saving results as pickles"""
    with open(f"train_result.pkl", "wb") as f:
        pickle.dump(results.history, f)
    with open(f"test_result.pkl", "wb") as f:
        pickle.dump(test_results, f)
        
    # del dataset, f, MODEL_FNAME_PATTERN, epochs, patience, start_time
    return results.history, test_results